const { getAbilityLevel } = require('../isomorphic-helpers');

const BattleCalc = {
  adjacent: function (a, b) {
    return BattleCalc.boardDistance(a, b) === 1;
  },

  tileBetweenThemIsOccupied: function (a, b, battleState) {
    if (BattleCalc.adjacent(a, b)) {
      return false;
    }

    var middle_x = (a.x + b.x) / 2;
    var middle_y = (a.y + b.y) / 2;

    var tile = battleState.getTileAtCoords(middle_x, middle_y);
    if (!tile) {
      return false;
    }

    return tile.occupied;
  },

  boardDistance: function (a, b) {
    if (!a || !b) {
      return NaN;
    }

    // const _xDif = a.x - b.x;
    // const _yDif = a.y - b.y;
    const _xDist = Math.abs(a.x - b.x); // (_xDif + (_xDif >> 31)) ^ (_xDif >> 31);
    const _yDist = Math.abs(a.y - b.y); //(_yDif + (_yDif >> 31)) ^ (_yDif >> 31);

    if (_xDist >= _yDist) {
      return _xDist;
    } else {
      return (_xDist + _yDist) * 0.5;
    }
  },

  getAlliesOfUnit: (unit, unitGroup, inclusive = true) => {
    if (typeof unitGroup === 'object') {
      unitGroup = Object.values(unitGroup);
    }

    unitGroup = unitGroup.filter((u) => !!u);

    return (
      inclusive
        ? unitGroup.filter((u) => u.team === unit.team)
        : unitGroup.filter((u) => u.team === unit.team && u.uid !== unit.uid)
    ).filter((u) => u.inPlay && !u.dead);
  },

  getNearestPiece: function (subject, pieces) {
    var nearest = pieces[0];
    var nearestDist = BattleCalc.boardDistance(subject, nearest);
    for (var i = 0; i < pieces.length; ++i) {
      if (BattleCalc.boardDistance(pieces[i], subject) < nearestDist) {
        nearest = pieces[i];
        nearestDist = BattleCalc.boardDistance(subject, nearest);
      }
    }
    return nearest;
  },

  getEquippedAbility: function (unit, abilityHandle) {
    if (typeof unit.equipped_abilities === 'object') {
      unit.equipped_abilities = Object.values(unit.equipped_abilities);
    }
    for (const handle of unit.equipped_abilities) {
      if (handle === abilityHandle) {
        if (typeof unit.abilityTrees === 'object') {
          unit.abilityTrees = Object.values(unit.abilityTrees);
        }
        for (const tree of unit.abilityTrees) {
          if (
            tree.handle === 'bow_use' &&
            unit?.equipment?.off_hand?.type !== 'bow'
          ) {
            continue;
          }

          for (const [handle, ability] of Object.entries(tree.abilities)) {
            if (handle === abilityHandle) {
              return ability;
            }
          }
        }
      }
    }
  },

  getPassiveAbility: function (unit, abilityHandle) {
    if (typeof unit.abilityTrees === 'object') {
      unit.abilityTrees = Object.values(unit.abilityTrees);
    }
    for (const tree of unit.abilityTrees) {
      for (const [handle, ability] of Object.entries(tree.abilities)) {
        if (handle === abilityHandle && ability.passive) {
          return ability;
        }
      }
    }
  },

  getLeapAPCostByDist: function (unit, dist) {
    const leap = BattleCalc.getEquippedAbility(unit, 'leap');
    if (leap) {
      return (unit.max_ap / leap.range.max) * dist; //  leap.apCost[ dist-leap.range.min ];
    }
    return NaN;
  },

  getMaxNumSummons({ ability_handle, caster_build }) {
    const ability_level = getAbilityLevel({
      ability_handle,
      unit_build: caster_build,
    });
    switch (ability_level) {
      case 1:
        return 2;
      case 2:
        return 3;
      case 3:
        return 4;
      default:
        return 5;
    }
  },

  getTeleportMPCostByDist: function (unit, dist) {
    const teleport = BattleCalc.getEquippedAbility(unit, 'teleport');
    if (teleport) {
      return teleport.mpCost[dist];
    }
    return NaN;
  },

  getCondition: function (unit, conditionHandle) {
    for (const condition of Object.values(unit.conditions) || {}) {
      if (condition.handle === conditionHandle) {
        return condition;
      }
    }
    return null;
  },

  clearDistTwoVictims: function (actor, victims) {
    for (var i = 0; i < victims.length; ++i) {
      if (BattleCalc.boardDistance(actor, victims[i]) === 2) {
        victims.splice(i, 1);
      }
    }

    return victims;
  },

  getAttTotal(unit, attTag) {
    const atts = unit.attributes;
    return (
      atts.base[attTag] +
      atts.fromUpgrades[attTag] +
      atts.fromEquipment[attTag] +
      atts.fromConditions[attTag]
    );
  },

  projectileTime: function (tilesPerSecond, a, b) {
    var dist = BattleCalc.boardDistance(a, b);

    return (dist / tilesPerSecond) * 1000;
  },

  getCooldown: function (unit) {
    return Math.max(0.5, 1.5 - unit.attributes.speed / 30);
  },

  pieceIsBetweenPair: function (subject, pair) {
    if (pair.length !== 2) {
      throw new Error(
        'Invalid vector length in BattleCalc::pieceIsBetweenPairOfPieces()'
      );
    }

    var pairDist = BattleCalc.boardDistance(pair[0], pair[1]);
    var avgOfDistancesFromSubjToNodes =
      (BattleCalc.boardDistance(subject, pair[0]) +
        BattleCalc.boardDistance(subject, pair[1])) /
      2;

    return avgOfDistancesFromSubjToNodes <= pairDist / 2;
  },

  unitHPAbovePercent: function (unit, threshold) {
    return (unit.hp / unit.max_hp) * 100 > threshold;
  },

  unitHPBelowPercent: function (unit, threshold) {
    return (unit.hp / unit.max_hp) * 100 < threshold;
  },

  unitAPAbovePercent(unit, threshold) {
    return (unit.ap / unit.max_ap) * 100 > threshold;
  },

  rollForReflectMagic(defenderUnit) {
    const reflectAbility = BattleCalc.getPassiveAbility(
      defenderUnit,
      'reflect_magic'
    );
    if (Math.random() <= reflectAbility?.potency ?? 0) {
      return true;
    }
    return false;
  },
};
module.exports = BattleCalc;
