const BattleCalc = require('../BattleCalc');
const MoveTowardsDecision = require('./movement/MoveTowardsDecision');
const MoveAwayDecision = require('./movement/MoveAwayDecision');
const DPSCloseDecision = require('./dps_close/DPSCloseDecision');
const DPSRangedDecision = require('./dps_ranged/DPSRangedDecision');
const SupportDecision = require('./support/SupportDecision');
const InterferenceDecision = require('./interference/InterferenceDecision');
const TankDecision = require('./tank/TankDecision');
const { logError } = require('../logger');

const AIHandler = function (battleState) {
  var _unitDecisions = {};
  this.getUnitDecisions = () => {
    return _unitDecisions;
  };

  this.addUnitDecisions = (unit) => {
    try {
      if (!_unitDecisions) {
        return;
      }

      this.removeUnitDecisions(unit.uid);

      //  breakIncapacitation: new BreakIncapacitationDecision( unit, battleState ),
      //  getBehindTank: new GetBehindTankDecision( unit, battleState ),
      var moveTowards = new MoveTowardsDecision(unit, battleState);
      var moveAway = new MoveAwayDecision(unit, battleState);
      var dps_close = new DPSCloseDecision(unit, battleState, moveTowards);
      var dps_ranged = new DPSRangedDecision(unit, battleState);
      var interference = new InterferenceDecision(unit, battleState);
      var support = new SupportDecision(unit, battleState);
      var tank = new TankDecision(unit, battleState, dps_close);

      _unitDecisions[unit.uid] = {
        //  breakIncapacitation: new BreakIncapacitationDecision( unit, battleState ),
        //  getBehindTank: new GetBehindTankDecision( unit, battleState ),
        moveTowards,
        moveAway,
        dps_close,
        dps_ranged,
        interference,
        support,
        tank,
      };

      for (var i = 0; i < unit.ai.priorities.length; ++i) {
        _unitDecisions[unit.uid][i] =
          _unitDecisions[unit.uid][unit.ai.priorities[i]];
      }
    } catch (err) {
      logError(err, {
        module: 'AIHandler',
        func: 'addUnitDecisions',
        unit: JSON.stringify(unit),
      });
    }
  };

  this.removeUnitDecisions = (unitId) => {
    for (const decision of Object.values(_unitDecisions[unitId] || {})) {
      decision.dispose();
    }
    delete _unitDecisions[unitId];
  };

  for (const unit of Object.values(battleState.allUnits || {})) {
    this.addUnitDecisions(unit);
  }

  this.getAction = (unitId, blacklistedAbilities) => {
    if (!battleState || !battleState.allPieces) {
      return 'wait';
    }

    const actor = battleState.allPieces[unitId];
    let action = null;

    if (!actor || !actor.ai) {
      return 'wait';
    }

    if (battleState.getEnemiesOfUnit(actor).length === 0) {
      return 'wait';
    }

    var enraged = BattleCalc.getCondition(actor, 'enraged');
    if (enraged) {
      var rageTarget = battleState.allUnits[enraged.instigatorId];
      if (!rageTarget || rageTarget.dead || !rageTarget.inPlay) {
        battleState.expireCondition(enraged.uid);
      } else {
        return attemptAction({
          actorId: actor.uid,
          abilityHandle: 'melee_attack',
          victimId: enraged.instigatorId,
        });
      }
    }

    var scared = BattleCalc.getCondition(actor, 'scared');
    if (scared) {
      var fearSource = battleState.allUnits[scared.instigatorId];
      if (!fearSource || fearSource.dead || !fearSource.inPlay) {
        battleState.expireCondition(scared.uid);
      } else {
        action = _unitDecisions[unitId].moveAway.query(
          battleState.allUnits[scared.instigatorId]
        );
        if (action && !blacklistedAbilities.has(action.abilityHandle)) {
          return attemptAction(action);
        } else {
          return 'wait';
        }
      }
    }

    if (typeof actor.ai.priorities === 'object') {
      actor.ai.priorities = Object.values(actor.ai.priorities);
    }

    // in case any of my allies (including myself) is critically low on HP, let's try the support tree one time
    const allies = BattleCalc.getAlliesOfUnit(
      actor,
      battleState.allUnits,
      true
    );
    for (let i = 0; i < allies.length; ++i) {
      if (BattleCalc.unitHPBelowPercent(allies[i], 15)) {
        for (let j = 0; j < actor.ai.priorities.length; ++j) {
          if (
            actor.ai.priorities[j] === 'support' &&
            (BattleCalc.getEquippedAbility(actor, 'healing_prayer') ||
              BattleCalc.getEquippedAbility(actor, 'sanctuary') ||
              BattleCalc.getEquippedAbility(actor, 'holy_force') ||
              BattleCalc.getEquippedAbility(actor, 'suture'))
          ) {
            try {
              const decision = _unitDecisions[unitId];
              action = decision ? decision[j]?.query() : null;
            } catch (err) {
              logError(err, {
                module: 'AIHandler',
                func: 'attemptAction:critically_low_hp_ally',
                actor: JSON.stringify(actor),
              });
            }
            break;
          }
        }
        break;
      }
    }

    if (action && !blacklistedAbilities.has(action.abilityHandle)) {
      return attemptAction(action);
    }

    // as a default, go in priority order through our decision trees, looking for an action to take
    for (let i = 0; i < actor.ai.priorities.length; ++i) {
      // chance to skip each tree (not heroes) - we do this to break annoying loops & to help units behave more dynamically
      if (
        !actor.hero &&
        actor.ai.settings.chanceToSkipRoles &&
        Math.random() < 0.075
      ) {
        continue;
      }

      try {
        const decision = _unitDecisions[unitId];
        action = decision ? decision[actor.ai.priorities[i]].query() : null;

        if (action && !blacklistedAbilities.has(action.abilityHandle)) {
          return attemptAction(action);
        }
      } catch (err) {
        logError(err, {
          module: 'AIHandler',
          func: 'getAction_waterfall',
          action: JSON.stringify(action),
        });
      }
    }

    // finally fall back on the MeleeDPS tree (everyone has a melee attack)
    try {
      const decision = _unitDecisions[unitId];
      action = decision ? decision.dps_close.query() : null;
      if (action && !blacklistedAbilities.has(action.abilityHandle)) {
        return attemptAction(action);
      }
    } catch (err) {
      logError(err, {
        module: 'AIHandler',
        func: 'meleeDPS_fallback',
        action,
      });
    }

    // for some reason we failed to find any action whatsoever. fuck it.
    return 'wait';

    function attemptAction(action) {
      if (action) {
        var ability = BattleCalc.getEquippedAbility(
          actor,
          action.abilityHandle
        );

        var victim = battleState.allPieces[action.victimId];
        var dist = BattleCalc.boardDistance(actor, victim);

        if (
          dist > ability.range.max &&
          !BattleCalc.getCondition(actor, 'enraged')
        ) {
          action = _unitDecisions[unitId].moveTowards.query(victim);
          if (!action || blacklistedAbilities.has(action.abilityHandle)) {
            // i'm trying to get somewhere but movement is impossible (i.e. surrounded)
            const nearest_enemy = battleState.getNearestEnemyOf(actor);
            if (BattleCalc.adjacent(actor, nearest_enemy)) {
              action = _unitDecisions[unitId].dps_close.query(nearest_enemy);
            } else {
              return 'wait';
            }
          }

          ability = BattleCalc.getEquippedAbility(actor, action.abilityHandle);
        } else if (dist < ability.range.min) {
          action = _unitDecisions[unitId].moveAway.query(victim);
          if (!action || blacklistedAbilities.has(action.abilityHandle)) {
            if (BattleCalc.getCondition(actor, 'scared')) {
              return 'wait';
            } else {
              // i'm a ranged unit trying to get away from a melee unit, but i can't move away (e.g. surrounded)
              const nearest_enemy = battleState.getNearestEnemyOf(actor);
              if (BattleCalc.adjacent(actor, nearest_enemy)) {
                action = _unitDecisions[unitId].dps_close.query(nearest_enemy);
              } else {
                return 'wait';
              }
            }
          }
          ability = BattleCalc.getEquippedAbility(actor, action.abilityHandle);
        }

        var blessedFactor = 1.0;
        var blessed = BattleCalc.getCondition(actor, 'blessed');
        if (blessed) {
          blessedFactor -= blessed.potency / 100;
        }

        var apCost =
          (ability.handle === 'leap'
            ? BattleCalc.getLeapAPCostByDist(
                actor,
                BattleCalc.boardDistance(actor, victim)
              )
            : ability.apCost) * blessedFactor;
        var mpCost =
          (ability.handle === 'teleport'
            ? BattleCalc.getTeleportMPCostByDist(
                actor,
                BattleCalc.boardDistance(actor, victim)
              )
            : ability.mpCost) * blessedFactor;

        if (
          (!ability.apCost || actor.ap >= apCost) &&
          (!ability.mpCost || actor.mp >= mpCost)
        ) {
          return action;
        } else if (actor.ap < apCost) {
          var adrenaline = BattleCalc.getEquippedAbility(actor, 'adrenaline');
          if (adrenaline && actor.mp >= adrenaline.mpCost * blessedFactor) {
            return {
              actorId: unitId,
              abilityHandle: 'adrenaline',
              victimId: unitId,
            };
          } else if (actor.ai.settings.waitForAPMP) {
            return 'wait';
          }
        } else if (actor.mp < mpCost) {
          var mysticFocus = BattleCalc.getEquippedAbility(
            actor,
            'mystic_focus'
          );
          if (mysticFocus) {
            if (actor.ap >= mysticFocus.apCost * blessedFactor) {
              return {
                actorId: unitId,
                abilityHandle: 'mystic_focus',
                victimId: unitId,
              };
            } else if (actor.ai.settings.waitForAPMP) {
              return 'wait';
            }
          } else if (actor.ap >= actor.max_ap) {
            // TODO: attack something
            return 'wait';
          } else {
            return 'wait';
          }
        } else if (actor.ai.settings.waitForAPMP) {
          // TODO: try to heal, try to buff, try to attack anything
          return 'wait';
        }
      }

      // TODO: try to heal, try to buff, try to attack anything

      return 'wait';
    }
  };

  this.dispose = () => {
    for (const unitId in _unitDecisions) {
      for (const decision in _unitDecisions[unitId]) {
        _unitDecisions[unitId][decision].dispose();
      }
    }
    _unitDecisions = null;
  };
};
module.exports = AIHandler;
