'use strict';

const app_tier = process.env.DT_TIER || 'localdev';

const getNewRelicConfig = params => {
  const { app_name } = params;
  /**
   * New Relic agent configuration.
   *
   * See lib/config/default.js in the agent distribution for a more complete
   * description of configuration variables and their potential values.
   */
  return {
    /**
     * Whether the module is enabled.
     *
     * @env NEW_RELIC_ENABLED
     */
    agent_enabled: !!process.env.ENABLE_DT_LOGGING,

    /**
     * Array of application names.
     */
    app_name: [].concat(['dt', `dt-${app_tier}`], [`${app_name}-${app_tier}`]),
    /**
     * Your New Relic license key.
     */
    license_key: process.env.NEW_RELIC_LICENSE_KEY,
    /**
     * This setting controls distributed tracing.
     * Distributed tracing lets you see the path that a request takes through your
     * distributed system. Enabling distributed tracing changes the behavior of some
     * New Relic features, so carefully consult the transition guide before you enable
     * this feature: https://docs.newrelic.com/docs/transition-guide-distributed-tracing
     * Default is false.
     */
    distributed_tracing: {
      /**
       * Enables/disables distributed tracing.
       *
       * @env NEW_RELIC_DISTRIBUTED_TRACING_ENABLED
       */
      enabled: false,
    },
    logging: {
      /**
       * Level at which to log. 'trace' is most useful to New Relic when diagnosing
       * issues with the agent, 'info' and higher will impose the least overhead on
       * production applications.
       */
      level: 'info',
    },
    /**
     * When true, all request headers except for those listed in attributes.exclude
     * will be captured for all traces, unless otherwise specified in a destination's
     * attributes include/exclude lists.
     */
    allow_all_headers: true,
    attributes: {
      /**
       * Prefix of attributes to exclude from all destinations. Allows * as wildcard
       * at end.
       *
       * NOTE: If excluding headers, they must be in camelCase form to be filtered.
       *
       * @env NEW_RELIC_ATTRIBUTES_EXCLUDE
       */
      exclude: [
        'request.headers.cookie',
        'request.headers.authorization',
        'request.headers.proxyAuthorization',
        'request.headers.setCookie*',
        'request.headers.x*',
        'response.headers.cookie',
        'response.headers.authorization',
        'response.headers.proxyAuthorization',
        'response.headers.setCookie*',
        'response.headers.x*',
      ],
    },
  };
};

const logError = (err, customAttributes) => {
  console.error(err);
  if (process.env.ENABLE_DT_LOGGING) {
    newrelic.noticeError(err, customAttributes);
  }
};

const logEvent = (eventType, attributes) => {
  newrelic.recordCustomEvent(eventType, attributes);
};

module.exports = {
  getNewRelicConfig,
  logError,
  logEvent,
};

// don't ask me why, but requiring this at the beginning of the file causes this module to export an empty obj
const newrelic = require('newrelic');
